import os
from epyseg.settings.global_settings import set_UI # set the UI to be used py qtpy
set_UI()
os.environ['SM_FRAMEWORK'] = 'tf.keras'  # set env var for changing the segmentation_model framework
import sys
from epyseg.uitools.blinker import Blinker
import logging
import traceback
from epyseg.deeplearning.deepl import EZDeepLearning
from qtpy.QtWidgets import QListWidgetItem, QAbstractItemView, QSpinBox, QComboBox, QProgressBar, \
    QVBoxLayout, QHBoxLayout, QLabel, QCheckBox, QRadioButton, QButtonGroup, QGroupBox, \
    QTextBrowser, QToolTip, QDoubleSpinBox
from qtpy.QtCore import Qt, QThreadPool, QPoint, QRect
from qtpy.QtGui import QColor, QTextCharFormat, QTextCursor, QPixmap, QIcon
from qtpy.QtWidgets import QGridLayout, QListWidget, QFrame, QTabWidget
from epyseg.gui.open import OpenFileOrFolderWidget
from qtpy.QtWidgets import QApplication
from qtpy import QtWidgets, QtCore, QtGui
from epyseg.tools.qthandler import XStream, QtHandler
from epyseg.gui.pyqtmarkdown import PyQT_markdown
# from epyseg.img import Img
from qtpy.QtWidgets import QPushButton, QWidget
from epyseg.tools.logger import TA_logger # logging

logger = TA_logger()

from qtpy.QtCore import PYQT_VERSION_STR
if PYQT_VERSION_STR<'6':
    # get code ready for pyqt6 where hi dpi is enabled by default
    try:
        QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling, True)  # high DPI fix
    except:
        pass

DEBUG = True  # set to True if GUI crashes
__MAJOR__ = 0
__MINOR__ = 0
__MICRO__ = 1
__RELEASE__ = ''  # a #b  # https://www.python.org/dev/peps/pep-0440/#public-version-identifiers --> alpha beta, ...
__VERSION__ = ''.join([str(__MAJOR__), '.', str(__MINOR__), '.',
                       str(__MICRO__)])  # if __MICRO__ != 0 else '', __RELEASE__]) # bug here fix some day
__AUTHOR__ = 'Benoit Aigouy'
__NAME__ = 'Deep Tools'
__EMAIL__ = 'baigouy@gmail.com'

# do something to play with the model and save various layers
# do something to create the training files and or maybe to check things...
# TODO
# load/visualize filters or weights ??? --> can be useful...

# make a tool to load and edit models with and without weights --> just get the previous stuff
# do something that shows all the layers in a combo or tree so that one can save or show every layer --> hope this will not crash but probably ok...
# the tree is a good idea as I can collapse or open it and can select one or several layers --> TODO
# TODO create a tree out of the model
# TODO allow for block creation visually
# tree drag and drop may also be an easy way to build a model
# could also build by input output --> and connect like that --> good idea
# see how to show all the layers and to deal with them

# TODO maye use that https://www.tensorflow.org/tensorboard/graphs --> rather than reinventing the wheel...

# The function to be traced.
# @tf.function
# def my_func(x, y):
#   # A simple hand-rolled layer.
#   return tf.nn.relu(tf.matmul(x, y))
#
# # Set up logging.
# stamp = datetime.now().strftime("%Y%m%d-%H%M%S")
# logdir = 'logs/func/%s' % stamp
# writer = tf.summary.create_file_writer(logdir)
#
# # Sample data for your function.
# x = tf.random.uniform((3, 3))
# y = tf.random.uniform((3, 3))
#
# # Bracket the function call with
# # tf.summary.trace_on() and tf.summary.trace_export().
# tf.summary.trace_on(graph=True, profiler=True)
# # Call only one tf.function when tracing.
# z = my_func(x, y)
# with writer.as_default():
#   tf.summary.trace_export(
#       name="my_func_trace",
#       step=0,
#       profiler_outdir=logdir)

# https://stackoverflow.com/questions/43702323/how-to-load-only-specific-weights-on-keras --> load specific weights --> easy I guess
# https://www.tensorflow.org/resources/tools --> really cool too


# TODO --> something like that in fact
# model.save(filepath) method:
#
# custom_model = keras.models.Model(inputs=model.layers[0].input,
#                                   outputs=model.layers[-2].output)
#
# custom_model.save('model.h5')

class DeepTools(QWidget):
    '''a deep learning GUI

    '''

    def __init__(self, parent=None):
        '''init for gui

        Parameters
        ----------
        parent : QWidget
            parent window if any (for future dvpt such as tyssue analyzer)

        '''
        self.currently_selected_metrics = []
        super().__init__(parent)
        self.blinker = Blinker()
        self.initUI()
        self.to_blink_after_worker_execution = None
        self.deepTA = EZDeepLearning()  # init empty model

    def initUI(self):
        print('here')
        '''init ui

        '''
        self.threading_enabled = True
        # logging
        self.logger_console = QTextBrowser(self)
        self.logger_console.setReadOnly(True)
        self.logger_console.textCursor().movePosition(QTextCursor.Left, QTextCursor.KeepAnchor, 1)
        self.logger_console.setHtml('<html>')
        self.logger_console.ensureCursorVisible()
        self.logger_console.document().setMaximumBlockCount(1000)  # limits to a 1000 entrees
        if not DEBUG:
            XStream.stdout().messageWritten.connect(self.set_html_black)
            XStream.stderr().messageWritten.connect(self.set_html_red)
        self.handler = QtHandler()
        self.handler.setFormatter(logging.Formatter(TA_logger.default_format))
        # change default handler for logging
        TA_logger.setHandler(self.handler)

        self.setWindowTitle(__NAME__ + ' v' + str(__VERSION__))

        # Initialize tab screen
        self.tabs = QTabWidget(self)
        self.model_tab = QWidget()
        self.train_tab = QWidget()
        self.ensemble_tab = QWidget()
        self.predict_tab = QWidget()
        self.advanced_tab = QWidget()
        # self.post_process = QWidget() # redundant with predict

        # Add tabs
        self.tabs.addTab(self.model_tab, 'Model') # split concatenate visualize model and or weights/ can I rename also things and allow for freeze of the layers too...
        self.tabs.addTab(self.train_tab, 'Build Train/GT Sets = Image concatenation')
        # self.tabs.addTab(self.predict_tab, 'Predict')
        self.tabs.addTab(self.ensemble_tab, 'Ensemble')  # To combine several outputs/models to improve seg quality
        self.tabs.addTab(self.advanced_tab, 'Advanced')  # mutate model, feeze, 3D, ... TODO
        self.advanced_tab.setVisible(False)
        # self.tabs.setVisible(False)
        # self.tabs.addTab(self.post_process, 'Post Process') # redundant with predict

        self.tabs.currentChanged.connect(self._onTabChange)

        # creating model tab
        self.model_tab.layout = QGridLayout()
        self.model_tab.layout.setAlignment(Qt.AlignTop)
        self.model_tab.layout.setColumnStretch(0, 25)
        self.model_tab.layout.setColumnStretch(1, 75)
        self.model_tab.layout.setHorizontalSpacing(3)
        self.model_tab.layout.setVerticalSpacing(3)


        # should I allow this to be stored in config file, probably not...
        self.input_model = OpenFileOrFolderWidget(parent_window=self, label_text='Load model',is_file=True,
                                                  extensions="Supported Files (*.h5 *.H5 *.hdf5 *.HDF5 *.json *.JSON *.model);;All Files (*)",
                                                  tip_text='Drag and drop a model file here')

        self.input_weights = OpenFileOrFolderWidget(parent_window=self, label_text='Load weights (optional)',
                                                    is_file=True,
                                                    extensions="Supported Files (*.h5 *.H5 *.hdf5 *.HDF5);;All Files (*)",
                                                    tip_text='Drag and drop a single weight file here')  # TODO shall i add *.model ???

        # parameters for the pretrained models
        self.groupBox_pretrain = QGroupBox('Model',objectName='groupBox_pretrain')
        self.groupBox_pretrain.setEnabled(True)
        # groupBox layout
        self.groupBox_pretrain_layout = QGridLayout()
        self.groupBox_pretrain_layout.setAlignment(Qt.AlignTop)
        self.groupBox_pretrain_layout.setColumnStretch(0, 25)
        self.groupBox_pretrain_layout.setColumnStretch(1, 25)
        self.groupBox_pretrain_layout.setColumnStretch(2, 50)
        self.groupBox_pretrain_layout.setColumnStretch(3, 2)
        self.groupBox_pretrain_layout.setHorizontalSpacing(3)
        self.groupBox_pretrain_layout.setVerticalSpacing(3)

        self.groupBox_pretrain_layout.addWidget(self.input_model, 1, 0, 1, 3)
        self.groupBox_pretrain_layout.addWidget(self.input_weights, 2, 0, 1, 3)

        self.groupBox_pretrain.setLayout(self.groupBox_pretrain_layout)
        self.model_tab.layout.addWidget(self.groupBox_pretrain, 1, 0, 1, 2)
        self.model_tab.setLayout(self.model_tab.layout)

        # widget global layout
        table_widget_layout = QVBoxLayout()
        table_widget_layout.setAlignment(Qt.AlignTop)
        table_widget_layout.addWidget(self.tabs)

        log_and_main_layout = QHBoxLayout()
        log_and_main_layout.setAlignment(Qt.AlignTop)

        # TODO put this in a group to get the stuff
        log_groupBox = QGroupBox('Log',objectName='log_groupBox')
        log_groupBox.setEnabled(True)

        help = PyQT_markdown()
        try:
            # this_dir, this_filename = os.path.split(__file__)
            # help.set_markdown_from_file(os.path.join(this_dir, 'deeplearning/docs', 'getting_started2.md'),
            #                             title='getting started: predict using pre-trained network')
            # help.set_markdown_from_file(os.path.join(this_dir, 'deeplearning/docs', 'getting_started.md'),
            #                             title='getting started: build and train a custom network')
            # help.set_markdown_from_file(os.path.join(this_dir, 'deeplearning/docs', 'getting_started3.md'),
            #                             title='getting started: further train the EPySeg model on your data')
            # help.set_markdown_from_file(os.path.join(this_dir, 'deeplearning/docs', 'pretrained_model.md'),
            #                             title='Load a pre-trained model')
            # help.set_markdown_from_file(os.path.join(this_dir, 'deeplearning/docs', 'model.md'),
            #                             title='Build a model from scratch')
            # help.set_markdown_from_file(os.path.join(this_dir, 'deeplearning/docs', 'load_model.md'),
            #                             title='Load a model')
            # help.set_markdown_from_file(os.path.join(this_dir, 'deeplearning/docs', 'train.md'), title='Train')
            # help.set_markdown_from_file(os.path.join(this_dir, 'deeplearning/docs', 'predict.md'), title='Predict')
            # help.set_markdown_from_file(os.path.join(this_dir, 'deeplearning/docs', 'preprocessing.md'),
            #                             title='Training dataset parameters')
            # help.set_markdown_from_file(os.path.join(this_dir, 'deeplearning/docs', 'data_augmentation.md'),
            #                             title='Data augmentation')
            # TODO prevent this window to be resize upon tab changes
            pass
        except:
            traceback.print_exc()

        # Initialize tab screen
        self.help_tabs = QTabWidget(self)
        self.help_tabs.setMinimumWidth(500)
        self.log_tab = QWidget()
        self.help_html_tab = QWidget()
        self.settings_GUI = QWidget()

        # Add tabs
        self.help_tabs.addTab(self.log_tab, 'Log')
        self.help_tabs.addTab(self.help_html_tab, 'Help')
        self.help_tabs.addTab(self.settings_GUI, 'GUI settings')

        # creating model tab
        self.log_tab.layout = QVBoxLayout()
        self.log_tab.layout.setAlignment(Qt.AlignTop)

        # global system progress bar
        self.pbar = QProgressBar(self)

        self.log_tab.layout.addWidget(self.logger_console)
        # self.log_tab.layout.addWidget(self.instant_help)
        self.log_tab.layout.addWidget(self.pbar)
        self.log_tab.setLayout(self.log_tab.layout)

        self.help_html_tab.layout = QVBoxLayout()
        self.help_html_tab.layout.setAlignment(Qt.AlignTop)
        self.help_html_tab.layout.setContentsMargins(0, 0, 0, 0)
        self.help_html_tab.layout.addWidget(help)
        self.help_html_tab.setLayout(self.help_html_tab.layout)

        self.enable_threading_check = QCheckBox('Threading enable/disable',objectName='enable_threading_check')
        self.enable_threading_check.setChecked(True)
        # self.enable_threading_check.stateChanged.connect(self._set_threading)
        self.enable_debug = QCheckBox('Debug mode', objectName='enable_debug')
        self.enable_debug.setChecked(False)
        self.enable_debug.stateChanged.connect(self._enable_debug)

        self.settings_GUI.layout = QVBoxLayout()
        self.settings_GUI.layout.setAlignment(Qt.AlignTop)
        self.settings_GUI.layout.addWidget(self.enable_threading_check)
        self.settings_GUI.layout.addWidget(self.enable_debug)
        self.settings_GUI.setLayout(self.settings_GUI.layout)

        log_and_main_layout.addLayout(table_widget_layout)
        log_and_main_layout.addWidget(self.help_tabs)

        self.setLayout(log_and_main_layout)

        try:
            try:
                screen = QApplication.desktop().screenNumber(QApplication.desktop().cursor().pos())
                centerPoint = QApplication.desktop().screenGeometry(screen).center()
            except:
                from qtpy.QtGui import QGuiApplication
                centerPoint = QGuiApplication.primaryScreen().geometry().center()
            self.setGeometry(QtCore.QRect(centerPoint.x() - self.width(), centerPoint.y() - self.height(), self.width(),
                                          self.height()))
        except:
            pass

        # # monitor mouse position to show helpful tips/guide the user
        # self.setMouseTracking(True) # does not work well because of contained objects capturing mouse --> maybe simplest is to have the
        self.show()

    def _onTabChange(self):
        pass

    def set_html_red(self, text):
        # quick n dirty log coloring --> improve when I have time
        textCursor = self.logger_console.textCursor()
        textCursor.movePosition(QTextCursor.End)
        self.logger_console.setTextCursor(textCursor)
        format = QTextCharFormat()
        format.setForeground(QColor(255, 0, 0))  # red
        self.logger_console.setCurrentCharFormat(format)
        self.logger_console.insertPlainText(text)
        self.logger_console.verticalScrollBar().setValue(self.logger_console.verticalScrollBar().maximum())

    def set_html_black(self, text):
        # quick n dirty log coloring --> improve when I have time
        textCursor = self.logger_console.textCursor()
        textCursor.movePosition(QTextCursor.End)
        self.logger_console.setTextCursor(textCursor)
        format = QTextCharFormat()
        format.setForeground(QColor(0, 0, 0))  # black
        self.logger_console.setCurrentCharFormat(format)
        self.logger_console.insertPlainText(text)

    def _enable_debug(self):
        if self.enable_debug.isChecked():
            # enable debug extensive log
            logger.setLevel(TA_logger.DEBUG)
            logger.debug('Debug enabled...')
        else:
            # disable debug log
            logger.setLevel(TA_logger.DEFAULT)
            logger.debug('Debug disabled...')

if __name__ == '__main__':
    app = QApplication(sys.argv)
    w = DeepTools()
    w.show()
    sys.exit(app.exec_())



# faire un GT creator --> just takes a ROI of the image with the mask and store it in a folder
# --> maybe do that --> not that hard to do

# ça marche avec les recursions --> c'est vraiment parfait maintenant
# idealement faudrait voir si je peux augmenter l'execution du reassignment mais entre le warping et la swap correction c'est parfait et pas de big swapping --> really good in fact!!!
# juste finalizer ce code


# TODO try the matching either based on translation using mermaid warp or on overlap between the two images --> really try the two scenarios
# maybe also use that to identify segmentation errors
# maybe try to run the wshed locally
# try a dynamic fixer of the thing

# --> give it a try to all !!! --> TODO
# maybe that is the game changer --> TODO
# still use the first image of pyTA epyseg to compute the wshed of the image --> TODO
# and could use the seeds from the stuff


# really try both configs --> I'm almost done!!!

# code that at home
# flip my flies

# --> get the warped file and get the stuff
# identify new cells !!!

# can I transform waping into a velocity stuff
# read the warped image and try to match into onto the new image --> if that works then I'm done
#

# open the mask
# open the warped mask of the other or the warped tracks and get the mapping
# how can I convert the phi file to translations --> if I can do that then all will be easy
# think about it --> TODO
#

# need get the mask
# def

# translation matrix can easily be obtained from the warped segmentation mask and the corresponding image
# --> apply same translation to all pixels with the same color ???
# need think about that
# need convert phi to translation
# remove white and compute translation
# need apply warp everytime --> just try if can be loaded or not
import os.path
import traceback

import matplotlib.pyplot as plt
from skimage import measure
from skimage.measure import regionprops
import numpy as np
from epyseg.img import Img, RGB_to_int24, int24_to_RGB
from timeit import default_timer as timer

# if the stuff has been done by the module then
from epyseg.ta.colors.colorgen import get_forbidden_colors_int24
from epyseg.ta.tracking.registration import pre_register_images, apply_translation
from epyseg.ta.tracking.tools import assign_random_ID_to_missing_cells, get_TA_file, first_image_tracking
from epyseg.ta.tracking.tracking_error_detector_and_fixer import find_vertices, associate_cell_to_its_neighbors2, \
    associate_cells_to_neighbors_ID_in_dict, compute_neighbor_score, optimize_score, get_cells_in_image_n_fisrt_pixel, \
    map_track_id_to_label, apply_color_to_labels

# TODO also finalize this tracking code and offer it as a registration
# also offer it as a check for segmentation --> try run the wshed locally with that

# TODO --> if two or more cells match then maybe try to find the best fit
# --> a good idea --> maybe could use area or cell elongation as the main criterion


# def sort_rps_by_area(rps):
# --> TODO add the possibility to warp
# also do a code to reconnect tracks that I may also use to MT the tracking --> I may need to know how to handle shared colored cell list or do two by two and do a code that can handle the junction of al the two by two pairs --> TODO
# indeed try to do that
# hack the other tracking code so that it does not need the cells
# TODO add checks...
# I still need to have and store all the cells to avoid duplicating IDs --> also need check that this is done properly ...

# le code doit surement etre ok --> à finaliser... et à tester
# en fait le code peut aussi etre utilisé pour de la segmentation  --> à essayer...
# --> warper une segmentation existante et l'appliquer au suivant et si une autre seg est dispo alors voir si on peut prendre le meilleur des deux mondes...

# TODO --> also finalize the code so that it can be used on a regular image to track
# if first image --> need generate the iamge then I need to be sure whether the stuff is working or not
# I need to have the already assigned colors passed in too
from epyseg.tools.early_stopper_class import early_stop
from epyseg.utils.loadlist import loadlist


# TODO --> need pre register
def match_by_max_overlap(name_t1, name_t0, channel_of_interest=None,assigned_IDs=[], recursive_assignment=True, warp_using_mermaid_if_map_is_available=True, pre_register=True):
    # either compute the translation matrix ->
    # not so hard
    # do the code to warp mask or even better --> warp the

    # handcorrection is always there --> maybe use that
    # do not get the file but the stuff to match

    filename1_without_ext = os.path.splitext(name_t1)[0]

    # print('loading', os.path.join(filename1_without_ext, 'handCorrection.tif'))
    mask_t1 = Img(os.path.join(filename1_without_ext, 'handCorrection.tif'))
    # print('loading', os.path.join(os.path.splitext(int_24_warped_track_t0)[0], 'tracked_cells_resized.tif'))
    int_24_warped_track_t0 = RGB_to_int24(Img(os.path.join(os.path.splitext(name_t0)[0], 'tracked_cells_resized.tif')))

    trans_dim_0= trans_dim_1 = 0
    if warp_using_mermaid_if_map_is_available:
        try:
            from personal.mermaid.deep_warping_uing_mermaid_minimal import warp_image_directly_using_phi
            filename0_without_ext = os.path.splitext(name_t0)[0]
            if os.path.exists(os.path.join(filename0_without_ext, 'mermaid_map.tif')):
                warp_map = Img(os.path.join(filename0_without_ext, 'mermaid_map.tif'))
                int_24_warped_track_t0 = warp_image_directly_using_phi(int_24_warped_track_t0, warp_map,pre_registration=name_t0 if pre_register else None).astype(np.uint32)
            else:
                if pre_register:
                    trans_dim_0, trans_dim_1 = _pre_reg(name_t0, name_t1, channel_of_interest)
                    int_24_warped_track_t0 = apply_translation(int_24_warped_track_t0, -trans_dim_0, -trans_dim_1)
        except:
            # traceback.print_exc()
            print("Mermaid registration failed... continuing without")
            if pre_register:
                trans_dim_0, trans_dim_1 = _pre_reg(name_t0, name_t1, channel_of_interest)
                int_24_warped_track_t0 = apply_translation(int_24_warped_track_t0, -trans_dim_0, -trans_dim_1)
    else:
        if pre_register:
            trans_dim_0, trans_dim_1 = _pre_reg(name_t0, name_t1, channel_of_interest)
            int_24_warped_track_t0 = apply_translation(int_24_warped_track_t0, -trans_dim_0, -trans_dim_1)

    # print('trans_dim_0,trans_dim_1' ,trans_dim_0,trans_dim_1 )
    # if not warped --> try warping
    # use mermaid map if available

    if len(mask_t1.shape) == 3:
        mask_t1 = mask_t1[..., 0]

    label_t1 = measure.label(mask_t1, connectivity=1, background=255)

    rps_label_t1 = regionprops(label_t1)

    track_t1 = np.zeros_like(mask_t1, dtype=np.uint64)
    # maybe sort by size the rps

    # otherwise can I put
    # the thing is my swapper may fix the errors if I let the cells empty
    # shall I try to find best fit --> could also do it by neighborhood
    # think about it --> maybe try
    # otherwise take the cell with the closest area

    matched_IDs = []
    # TODO --> sort by size
    # KEEP we sort the rps by area directly
    for rps in sorted(rps_label_t1, key=lambda r: r.area, reverse=True): # this will create a bug if two cells have the same area ???
        # print('area',rps.area)

        # get coords and count unique colors in this guy
        # assign new cells

        # prevent assigning a cell twice, otherwise keep the two possibilities and find best choice
        if track_t1[rps.coords[0][0], rps.coords[0][1]] != 0:
            continue

        pixels = int_24_warped_track_t0[rps.coords[:, 0], rps.coords[:, 1]]
        # need get the most frequent from that and assign it if not already assigned
        # maybe sorting cells by size would be a good idea ???
        unique, counts = np.unique(pixels, return_counts=True)
        # print(unique, counts)

        # assign the found ID
        color_to_assign = unique[np.argmax(counts)]

        # prevent assigning the ID --> could also remove the parent cell
        if color_to_assign in matched_IDs:
            continue

        matched_IDs.append(color_to_assign)

        track_t1[rps.coords[:, 0], rps.coords[:, 1]] = color_to_assign

    # could get the centroid and apply it as a seed to for better segmentation !!!
    # I can also map cells backward and forward to see errors, ideally I should keep

    # Img(int24_to_RGB(track_t_cur), dimensions='hwc').save(TA_path_plus_1+"/tracked_cells_resized_swap.tif", mode='raw')
    # Img(int24_to_RGB(track_t1), dimensions='hwc').save('/E/Sample_images/sample_images_PA/trash_test_mem/trash_registration_network/1/test_track_using_warp.tif', mode='raw')

    # here !=


    # get matched cells
    # matched_cells_ids = get_matched_ids(rps_label_t1, track_t1)

    # print('sume here', np.sum(track_t1-int_24_warped_track_t0))

    # track_t1 = match_unmatched_cells(track_t1, int_24_warped_track_t0, matched_cells_ids, label_t1, rps_label_t1,
    #                                  assigned_IDs)

    track_t1 = assign_random_ID_to_missing_cells(track_t1, label_t1, regprps=rps_label_t1, assigned_ids=assigned_IDs)

    # just for TA compatibility...
    track_t1[label_t1 == 0] = 0xFFFFFF



    # plt.imshow(int24_to_RGB(tracks)-int24_to_RGB(int_24_warped_track_t0)) # both are the same
    # plt.imshow(int24_to_RGB(track_t1)) # both are the same
    # plt.show()

    # print('sume here', np.sum(track_t1-int_24_warped_track_t0))
    # then I would need to add the cells and handle the possible swapped cells --> see how to do that in the best and fastest way
    # for sur this algo will be much faster than the one with
    # can i do it recursively ??? --> apply registration recursively --> if so will it manage to perfectly align the stuff --> ????

    # all quite good need handle dupes --> maybe take the closest or the one with the most homogeneous translation
    # can also be combined with the swapping correscion

    # print(track_t1.shape)
    # print(int24_to_RGB(track_t1).shape)

    # ideally I should preserve the neighborhood too

    # TODO --> need add the missing
    # see how I can use the local IDs as labels ???
    # just copy big chunks of code from my
    # Img(int24_to_RGB(track_t1), dimensions='hwc').save('/E/Sample_images/sample_images_PA/trash_test_mem/trash_registration_network/1/test_track_using_warp.tif', mode='raw')

    if recursive_assignment:
        # TODO

        # print('sume here', np.sum(track_t1-int_24_warped_track_t0)) # --> 0 --=> identical images --> but WHY???

        # Img(int24_to_RGB(track_t1), dimensions='hwc').save('/E/Sample_images/sample_images_PA/trash_test_mem/trash_registration_network/1/zoubui1.tif', mode='raw')
        # Img(int24_to_RGB(int_24_warped_track_t0), dimensions='hwc').save('/E/Sample_images/sample_images_PA/trash_test_mem/trash_registration_network/1/zoubui2.tif', mode='raw')

        # print(track_t1.shape, int_24_warped_track_t0.shape, track_t1.dtype, int_24_warped_track_t0.dtype) # --> 2048 * 2048 uint64 uint32 --> why uint32 --> is that realy what I want by the way???
        #.astype(np.uint64)

        # why I don't have my cells there --> why no error --> do I have a big error in my code --> that is possible...

        # the two masks are identical --> there is a big bug!!!

        run_swapping_correction_recursively_to_further_identify_cells(track_t1, int_24_warped_track_t0, label_t1,
                                                                      rps_label_t1, assigned_IDs, filename1_without_ext,
                                                                      MAX_ITER=15)
    else:
        Img(int24_to_RGB(track_t1)).save(get_TA_file(filename1_without_ext, 'tracked_cells_resized.tif'), mode='raw') # save directly...


def _pre_reg(name_t0, name_t1, channel_of_interest):

        I0 = Img(name_t0)
        if channel_of_interest is not None and len(I0.shape) > 2:
            I0 = I0[..., channel_of_interest]
        I1 = Img(name_t1)
        if channel_of_interest is not None and len(I1.shape) > 2:
            I1 = I1[..., channel_of_interest]
        trans_dim_0, trans_dim_1 = pre_register_images(orig_t0=I0, orig_t1=I1)
        return trans_dim_0, trans_dim_1


def get_matched_ids(rps_t1_mask, tracks):
    matched_cells_ids = []

    for iii, region in enumerate(rps_t1_mask):
        color = tracks[region.coords[0][0], region.coords[0][1]]
        if color == 0:
            # missing_cells_t1.append(iii)
            pass
        else:
            matched_cells_ids.append(color)

    return matched_cells_ids


# def match_unmatched_cells(tracks, tracked_cells_t0, matched_cells_ids, labels_t1, rps_t1_mask, assigned_ids):
#     # now we handle unmatched cells
#
#     labels_tracking_t0 = measure.label(tracked_cells_t0, connectivity=1, background=0xFFFFFF)
#
#     # plt.imshow(tracked_cells_t0[...,0], cmap='gray') #
#     # plt.show()
#
#     rps_t0 = regionprops(labels_tracking_t0)
#
#     unmatched_cells_in_t0 = []
#     # all ok so why bug there
#     for iii, region in enumerate(rps_t0):
#         # there is a bug there and I don't get it
#         color = tracked_cells_t0[region.coords[0][0], region.coords[0][1]]
#         # print(color)  # why always the same color -->
#         if color not in matched_cells_ids:  # why bug --> is it due to hash error
#             unmatched_cells_in_t0.append(iii)
#
#     print('unmatched_cells_in_t0',unmatched_cells_in_t0) # no unmatched cell
#
#     tracks = assign_random_ID_to_missing_cells(tracks, labels_t1, regprps=rps_t1_mask, assigned_ids=assigned_ids)
#
#     # just for TA compatibility...
#     tracks[labels_t1 == 0] = 0xFFFFFF
#
#     return tracks


# handle first image and all possible pairs --> TODO


# TODO make this step an option in the code but probably ok
# TODO --> do that and call it -> probably not that hard
# all is ok
def run_swapping_correction_recursively_to_further_identify_cells(tracks, tracked_cells_t0, labels_t1, rps_t1_mask,
                                                                  assigned_ids, filename1_without_ext, MAX_ITER=15):
    start_all = timer()
    # print('intermediate time before swap correction0', timer() - start_loop)

    # THIS IS STARTING FROM HERE THAT I SHOULD USE THE NEW SCORING AND SWAPPING ALGO --> TODO
    # Img(int24_to_RGB(tracks)).save(get_TA_file(filename1_without_ext, 'tracked_cells_resized.tif'),mode='raw')

    ##################################################################################NEO CODE

    # en fait c'est très bon j'ai juste à peaufiner pr sauver du temps de processing et aussi ajouter le score --> avec un nb max de recursion
    # prevent analysing too many times the images

    # get score and if not imrpoving anymore --> early stop and if improves get a max nb of recursion and really speed up the whole process
    # see how to better handle missing cells without reprocessing everything !!!



    # just for mapping
    track_t_cur = tracks
    track_t_minus_1 = tracked_cells_t0

    # recursion is really needed but I don't need do all that for the recursion
    # assign new cell

    # could compare this to the other
    # vertices_t_minus_1 = find_vertices(track_t_minus_1, return_vertices=True, return_bonds=False)
    vertices_t_minus_1 = np.stack(np.where(track_t_minus_1 == 0xFFFFFF), axis=1)
    # vertices_t_cur = find_vertices(track_t_cur, return_vertices=True, return_bonds=False)
    vertices_t_cur = np.stack(np.where(track_t_cur == 0xFFFFFF), axis=1)

    # print('vertices_t_minus_1',vertices_t_minus_1)
    # print('vertices_t_cur',vertices_t_cur)

    # print('sum all', np.sum(vertices_t_minus_1-vertices_t_cur))
    # somehow the two images are the same --> did I do a mistake somewhere ???

    # print(vertices_t_minus_1.shape)

    # plt.imshow(vertices_t_minus_1)
    # plt.show()

    # maybe also if too many rejected changes --> ignore
    # DO I NEED SO MUCH PYRAMIDAL REG WITH SUCH A POWERFUL ALGO ??? --> probably not ??? --> check that and really compare to the other method and clean code heavily !!!

    cells_and_their_neighbors_minus_1 = np.unique(
        np.asarray(associate_cell_to_its_neighbors2(vertices_t_minus_1, track_t_minus_1)), axis=0)
    cells_and_their_neighbors_minus_1 = associate_cells_to_neighbors_ID_in_dict(cells_and_their_neighbors_minus_1)


    # cells_in_t_minus_1 = np.unique(track_t_minus_1)
    # print('neighbs found', cells_and_their_neighbors_minus_1[10287288])


    # il y a des bugs dans ce code --> maybe recode all
    # some of the cells are not

    initial_score = -1

    # it's better but there are still some swapping --> need be fixed
    for lll in range(MAX_ITER):
        start_loop = timer()

        # print(get_vx_neighbors(vertices, RGB24img=img)) # is that slow??? --

        # get_vx_neighbors(vertices_t_minus_1,                     RGB24img=track_t_minus_1)  # ça marche et pas trop long par contre ça me donne juste la correspondance one to one avec le vx array ce qui est peut etre ce que je veux d'ailleurs mais faut pas changer l'ordre

        # ce truc est slow --> can I speed up using vertices --> most likely yes
        # print(np.unique(np.asarray(associate_cell_to_its_neighbors(img)), axis=0)) # really get cell neighbors # ça a l'air de marcher # maybe I can speed up by using just the vertices or not ???? --> think about it
        # print(associate_cell_to_its_neighbors(img)) # really get cell neighbors # ça a l'air de marcher # maybe I can speed up by using just the vertices or not ???? --> think about it
        # aussi voir comment associer les vertices à une seule cellule --> car très utile
        # print()  # a bit faster indeed but not outstanding

        # get cells and their neigbs for -1 and 0 and do a scoring --> will highlight potential swapping and or division and or other things in a way --> REALLY TRY THAT

        cells_and_their_neighbors_cur = np.unique(
            np.asarray(associate_cell_to_its_neighbors2(vertices_t_cur, track_t_cur)), axis=0)
        cells_and_their_neighbors_cur = associate_cells_to_neighbors_ID_in_dict(cells_and_their_neighbors_cur)


        # because this is the label of the cell and not its id ???
        # print('difference', set(cells_and_their_neighbors_cur.keys())-set(cells_and_their_neighbors_minus_1.keys())) # this difference is 0 --> but why
        # print('difference2', set(cells_and_their_neighbors_minus_1.keys())-set(cells_and_their_neighbors_cur.keys())) # this difference is 0 --> but why # both differences are 0 --> does not make any sense then
        # print('1',len(cells_and_their_neighbors_cur.keys())) # both exactly have the same size --> have I done a mistake ???
        # print('2',len(cells_and_their_neighbors_minus_1.keys()))





        # FOR DEBUG
        # print('cells_and_their_neighbors_minus_1', cells_and_their_neighbors_minus_1)
        # print('cells_and_their_neighbors_cur', cells_and_their_neighbors_cur)

        # print('neighbs found2', cells_and_their_neighbors_minus_1[10287288])
        # print('neighbs found', cells_and_their_neighbors_cur[10287288])
        # print('cells_and_their_neighbors_plus_1', cells_and_their_neighbors_plus_1)

        # for all the common cells --> compute a score
        # nb of matching neighbs / total nb of neighbs in both

        # create and color

        # cells_present_in_t_cur_but_absent_in_t_plus_1 = []
        # cells_present_in_t_cur_but_absent_in_t_plus_1_score={}
        # matching_neighorhood_score_t_cur_with_t_plus_1 = np.zeros_like(track_t_cur, dtype=float)
        #
        # compute_neighbor_score(cells_and_their_neighbors_cur, cells_and_their_neighbors_plus_1,cells_present_in_t_cur_but_absent_in_t_plus_1, cells_present_in_t_cur_but_absent_in_t_plus_1_score, matching_neighorhood_score_t_cur_with_t_plus_1)

        # for cell, neigbs in cells_and_their_neighbors_cur.items():
        #     if cell in cells_and_their_neighbors_plus_1:
        #         # compute score and color cell
        #         neigbsb = cells_and_their_neighbors_plus_1[cell]
        #         score = (len(list(set(neigbs) & set(neigbsb))) * 2) / (len(neigbsb) + len(neigbs))
        #         cells_present_in_t_cur_but_absent_in_t_plus_1_score[cell]=score
        #         matching_neighorhood_score_t_cur_with_t_plus_1[track_t_cur == cell] = score
        #     else:
        #         cells_present_in_t_cur_but_absent_in_t_plus_1.append(cell)
        #         cells_present_in_t_cur_but_absent_in_t_plus_1_score[cell] = 0

        cells_present_in_t_cur_but_absent_in_t_minus_1 = []
        cells_present_in_t_cur_but_absent_in_t_minus_1_score = {}

        # matching_neighorhood_score_t_cur_with_t_minus_1 = np.zeros_like(track_t_cur, dtype=float)

        score = compute_neighbor_score(cells_and_their_neighbors_cur, cells_and_their_neighbors_minus_1,
                                       cells_present_in_t_cur_but_absent_in_t_minus_1,
                                       cells_present_in_t_cur_but_absent_in_t_minus_1_score)#,score_plotting_image=matching_neighorhood_score_t_cur_with_t_minus_1)  # , cells_present_in_t_cur_but_absent_in_t_minus_1, cells_present_in_t_cur_but_absent_in_t_minus_1_score, matching_neighorhood_score_t_cur_with_t_minus_1)


        print('cells_present_in_t_cur_but_absent_in_t_minus_1', cells_present_in_t_cur_but_absent_in_t_minus_1) # --> big bug because this is empty and it does not make sense
        # plt.imshow(matching_neighorhood_score_t_cur_with_t_minus_1)
        # plt.show()

        # print('defined score', cells_present_in_t_cur_but_absent_in_t_minus_1_score[10287288])

        if initial_score < 0:
            initial_score = score
        # for cell, neigbs in cells_and_their_neighbors_cur.items():
        #     if cell in cells_and_their_neighbors_minus_1:
        #         # compute score and color cell
        #         neigbsb=cells_and_their_neighbors_minus_1[cell]
        #         score = (len(list(set(neigbs) & set(neigbsb)))*2)/(len(neigbsb)+len(neigbs))
        #         cells_present_in_t_cur_but_absent_in_t_minus_1_score[cell]=score
        #         matching_neighorhood_score_t_cur_with_t_minus_1[track_t_cur == cell]=score
        #     else:
        #         cells_present_in_t_cur_but_absent_in_t_minus_1.append(cell)
        #         cells_present_in_t_cur_but_absent_in_t_minus_1_score[cell] = 0

        # offer solutions based on scores --> TODO

        # do I have a bug here ???

        # print(type(cells_and_their_neighbors_cur), type(cells_and_their_neighbors_minus_1))  # le second est une liste mais ne devrait pas !!!! --> un bug qq part
        # test after optimization

        # can probably store that

        # ça rend le code un peu plus long mais vraiment pas deconnant --> doit etre assez facile à faire
        # lab_t_minus_1 = label(track_t_minus_1, connectivity=1, background=0xFFFFFF)
        # rps_t_minus_1 = regionprops(lab_t_minus_1)
        lab_t_cur = measure.label(track_t_cur, connectivity=1, background=0xFFFFFF)
        # rps_t_cur = regionprops(lab_t_cur)

        # can I read all the neighbors in a smart way now by first index of one

        # can compare neigbors for a cell in two instances --> give it a try
        # all should be easy I guess !!!

        # to get cells in one I can get unique of the first col and compare it to the other --> in fact not that hard I think
        # pb if a cell has no vertex it will be ignored
        # --> purely isolated cell

        # in fact I can simply do unique otherwise directly on the image

        cells_in_t_cur, first_pixels_t_cur = get_cells_in_image_n_fisrt_pixel(track_t_cur)  # np.unique(track_t_cur)
        map_tracks_n_label_t_cur = map_track_id_to_label(first_pixels_t_cur, track_t_cur, lab_t_cur)

        # print('first_pixels_t_cur', first_pixels_t_cur)  # this is the ravel index --> how can I convert it back ???
        # print('cells_in_t_cur', cells_in_t_cur)
        # print('map_tracks_n_label_t_cur', map_tracks_n_label_t_cur)

        # I'm ready to use this for the tracking of cells
        # just see how fast and efficient this is

        # l'ancien code est bien meilleur en fait --> voir pkoi parce que le code là est plus beau et plus logique --> essayer de comprendre pkoi moins bon en fait est -ce trop stringent au niveau des scores ????
        # essayer de comprendre le pb!!!!

        FIX_SWAPS = True
        # threshold = 1.0 # 0.5 # 1.0
        threshold = 0.8  # 0.76 # 0.5 # 1.0 # really required to be above 1 to cope with concurrent drifts of several cells


        # even though there are errors it really doe not find them
        if FIX_SWAPS:
            # DETECT SWAPS AND FIX THEM AND COMPUTE SCORE --> TODO also fix cell correspondance then --> really necessary!!!
            # we update both the cells and their local correspondance --> I think I have it --> just need few more crosses to identify more errors
            # can I use that as a tracking algo on top of the other with a minimization of the stuff and just do the coloring in the end --> TODO

            cells_and_their_neighbors_cur, map_tracks_n_label_t_cur, last_score_reached = optimize_score(
                cells_and_their_neighbors_cur, cells_and_their_neighbors_minus_1,
                cells_present_in_t_cur_but_absent_in_t_minus_1_score, map_tracks_n_label_t_cur, threshold=threshold)
            # print('map_tracks_n_label_t_cur2',map_tracks_n_label_t_cur2)
            # print('map_tracks_n_label_t_cur',map_tracks_n_label_t_cur)



            # style if improvment < 5 early stop
            # score improved from 617.7013813251964 to 692.6711356897581
            # score improved from 692.6711356897581 to 708.2236112791516
            # score improved from 708.2236112791516 to 711.5193129630884
            # score improved from 711.5193129630884 to 712.606217724993
            # score improved from 712.606217724993 to 713.5926725744915

            # max reached after 10 iter on the same just for fun it stopped at 6 and I couldn't detect any improvment --> ok in fact!!!
            # use a detector of divisions and of T1s
            # can also be used to detect intercalation maybe ???

            # could also have a minimum improvment and break if not really improving anymore
            # 1388
            # do i have a bug because no possible alignment
            if last_score_reached == initial_score:
                print(last_score_reached, initial_score)
                print('early stop', lll)
                break

            if last_score_reached > initial_score:
                print('score improved from', initial_score, 'to', last_score_reached)
                initial_score = last_score_reached
            else:
                print('score did not improve', initial_score)

            # depending on the corrections I apply I could also identify cells that have changed between -1 and 1 and that are missassigned cells in 3 and
            # could in fact try tracking with three images to be more efficient in fact directly because post correction will be time consuming in fact especially for swapped cells
            # but still i need that
            # do a version of the tracking using pyramidal reg that relies on that!!! --> TODO

            # possible assignment 10240303 6824721 0.6666666666666666 (156, 65, 47) (104, 35, 17) 0.6666666666666666  --> a dividing cell --> really a good peak
            # 0
            # possible assignment 10369245 10443693 0.5714285714285714 (158, 56, 221) (159, 91, 173) 0.5714285714285714 --> une misegmented cell en bas au milieu) et une cellule adjacente qui elle est bien placee mais je suppose que faire ce changment fera baisser le core et que donc ce changement sera ignoré...
            # 0
            # possible assignment 11560212 8271832 0.5714285714285714 (176, 101, 20) (126, 55, 216) 0.5714285714285714 --> une autre misegmented cell à gauche un peu apres le centre de l'image!!! et une cellule bien trackee à coté --> pareil faire le score et voir
            # 0
            # possible assignment 12248378 7043456 1.0 (186, 229, 58) (107, 121, 128) 1.0 --> mis tracked cell

            # cells_in_t_cur = cells_and_their_neighbors_cur.keys()

            # TO MANUALLY FIND A CELL
            # for jjj, missing in enumerate(cells_in_t_cur):
            #     if missing == 0xFFFFFF:
            #         continue
            #
            #     # if missing != r_g_b_to_rgb( 4,77,109):# for sure it's a swapped cell
            #     #     continue
            #
            #     # the cell 9829855 is not shown properly --> it has no score --> why ?? maybe because no vx ???
            #     if missing == 10369245:# 11560212
            #         label_id = map_tracks_n_label_t_cur[jjj][1]
            #         rps = rps_t_cur[label_id - 1]
            #         bbox = rps.bbox
            #         # bbox = [bbox[0] - 50, bbox[1] - 50, bbox[2] + 50, bbox[3] + 50]
            #         # Bounding box ``(min_row, min_col, max_row, max_col)
            #         # do a crop of the region of interest
            #
            #         print('bbox pos of cell to assign' ,bbox)

            # plot change

            # print('tests of alm', 12248378 in cells_and_their_neighbors_cur.keys(), 12248378 in cells_and_their_neighbors_minus_1.keys(), 7043456 in cells_and_their_neighbors_cur.keys(), 7043456 in cells_and_their_neighbors_minus_1.keys())
            # False False True True --> ok in fact

            # check why some of the cells are swapped --> especially because it cannot imrpove the code --> see why not detected and really need change this !!!
            #
            compute_neighbor_score(cells_and_their_neighbors_cur, cells_and_their_neighbors_minus_1,
                                   cells_present_in_t_cur_but_absent_in_t_minus_1,
                                   cells_present_in_t_cur_but_absent_in_t_minus_1_score)

            # FOR DEBUG KEEP
            # matching_neighorhood_score_t_cur_with_t_minus_1 = np.zeros_like(track_t_cur, dtype=float)
            # compute_neighbor_score(cells_and_their_neighbors_cur, cells_and_their_neighbors_minus_1,
            #                        cells_present_in_t_cur_but_absent_in_t_minus_1,
            #                        cells_present_in_t_cur_but_absent_in_t_minus_1_score,
            #                        matching_neighorhood_score_t_cur_with_t_minus_1, track_t_cur)
            #
            #
            # plt.imshow(matching_neighorhood_score_t_cur_with_t_minus_1)
            # plt.show()

            # if I really need swap then I also need

            # ça marche mais du coup je dois lui faire faire la coloration maintenant

            corrected_track = apply_color_to_labels(lab_t_cur, map_tracks_n_label_t_cur)
            # plt.imshow(int24_to_RGB(corrected_track))
            # plt.show()

            corrected_track = assign_random_ID_to_missing_cells(corrected_track, labels_t1, regprps=rps_t1_mask,
                                                                assigned_ids=assigned_ids)
            # update mask!!!
            track_t_cur = corrected_track

        ##################################################################################END NEO CODE

        # # now we really do the swapping the swapping correction --> I could reuse this code in the other
        # # MEGA TODO just to speed up things here just restrict the swapping to the problematic cells and new cells could also exclude cells that match full dark of main image --> indeed could gain some time when big translations
        # vertices_t0 = detect_vertices_and_bonds(labels_t0)
        # vertices_t0 = np.where(vertices_t0 == 255)
        # vertices_t1 = detect_vertices_and_bonds(labels_t1)
        # vertices_t1 = np.where(vertices_t1 == 255)
        #
        # # ça c'est facile mais faut juste looper sur les cellules problematiques --> ajouter une liste de cellules à checker et les supprimer si ok ??? peut etre --> à faire
        # print('intermediate time before swap correction1', timer() - start_loop)
        # for fix in range(swap_fix_recursions):
        #     if __DEBUG__:
        #         print('swapping fix recursion #', fix)
        #     # print(tracked_cells_t1.shape)
        #     tracked_cells_t1 = _fix_swapping_internal(tracked_cells_t0, tracked_cells_t1, vertices_t0=vertices_t0,
        #                                               vertices_t1=vertices_t1, MIN_CUT_OFF_SWAPPING=MIN_CUT_OFF_SWAPPING)
        #     # print('-->', tracked_cells_t1.shape)
        #     # Img(tracked_cells_t1).save(get_TA_file(filename0_without_ext, 'track_after_swap_correction'+ str(fix)+ '.tif'), mode='raw')

        # example of swapped cells
        # print('total time', timer() - start_all)  # --> 47 secs --> un peu lent qd meme

        # Img(tracked_cells_t1).save(get_TA_file(filename0_without_ext, 'track_after_swap_correction.tif'), mode='raw')
        print('end loop', timer() - start_loop)

    print(get_TA_file(filename1_without_ext, 'tracked_cells_resized.tif'))

    # NB it still contains small swapping errors I don't get --> why
    # swap first then assign and assign if not swapped --> check cell

    # celui qui m'interesse c'est le 2
    # Img(int24_to_RGB(corrected_track)).save(get_TA_file(filename1_without_ext, 'tracked_cells_resized_v2.tif'),mode='raw')
    Img(int24_to_RGB(track_t_cur)).save(get_TA_file(filename1_without_ext, 'tracked_cells_resized.tif'),mode='raw')



    # with a recursion it's very good actually and really is optimized

    # print(r_g_b_to_rgb(137,134,15))

    print('total time all recusrsions', timer() - start_all)
    # return track_t_cur

def track_first_frame(original_t0, assigned_ids, seed):
    filename0_without_ext = os.path.splitext(original_t0)[0]
    mask_t0 = Img(get_TA_file(filename0_without_ext, 'handCorrection.tif'))

    if mask_t0.has_c():
        mask_t0 = mask_t0[..., 0]
    labels_t0 = measure.label(mask_t0, connectivity=1, background=255)  # FOUR_CONNECTED
    tracked_cells_t0 = first_image_tracking(mask_t0, labels_t0, assigned_ids=assigned_ids, seed=seed)
    Img(int24_to_RGB(tracked_cells_t0), dimensions='hwc').save(
        get_TA_file(filename0_without_ext, 'tracked_cells_resized.tif'), mode='raw')  # should work --> DO the job

def match_by_max_overlap_lst(lst,  channel_of_interest=None, recursive_assignment=True, warp_using_mermaid_if_map_is_available=True, pre_register=True, progress_callback=None):
    start_all = timer()
    # pass it an entire list so that it can do the tracking and evrything like that.
    seed = 1  # always start tracking with same seed to have roughly the same color
    assigned_ids = get_forbidden_colors_int24()

    try:
        zipped_list = zip(lst, lst[1:])
    except:
        # assume list contains only one image --> return just the first image tracking
        track_first_frame(lst[0], assigned_ids, seed)
        return

    for iii, (original_t0, original_t1) in enumerate(zipped_list):
        try:
            if early_stop.stop == True:
                return
            if progress_callback is not None:
                progress_callback.emit((iii * 100) / len(zipped_list))
            else:
                print(str((iii * 100) / len(zipped_list) ) + '%')
        except:
            pass

        if iii == 0:
            # need create the track for the first image then recursively go on
            track_first_frame(original_t0, assigned_ids, seed)

        match_by_max_overlap(original_t1, original_t0, channel_of_interest=channel_of_interest, recursive_assignment=recursive_assignment, warp_using_mermaid_if_map_is_available=warp_using_mermaid_if_map_is_available, pre_register=pre_register)
    print('total time', timer() - start_all)

# faire un GT generator --> takes ROI of crops and does it

# TODO --> do a GUI here for that stuff
if __name__ == '__main__':
    import sys

    if True:
        lst = loadlist('/E/Sample_images/sample_images_PA/trash_test_mem/mini_empty/list.lst')
        # match_by_max_overlap_lst(lst, recursive_assignment=False, warp_using_mermaid_if_map_is_available=True) # --> 2.76 secs without recursion
        # match_by_max_overlap_lst(lst, recursive_assignment=True, warp_using_mermaid_if_map_is_available=True) # --> 21 secs with recursion vs 2.76 secs without recursion... --> *10

        # shall I assign or not lost cells
        # I should probably ask for prereg or not --> check
        # or do pre reg by default because fast and useful --> I would need the channel
        match_by_max_overlap_lst(lst, channel_of_interest=1, recursive_assignment=True, warp_using_mermaid_if_map_is_available=False, pre_register=True) # good and does not load mermaid files yet
        sys.exit(0)

    # if True:
    # ça marche meme si la key est dupliquée...
    #     student_tuples = [
    #         ('john', 'A', 15),
    #         ('jane', 'B', 12),
    #         ('dave', 'B', 10),
    #         ('dan', 'C', 10),
    #         ]
    #     print(sorted(student_tuples, key=lambda student: student[2]))  # sort by age
    #
    #     sys.exit(0)

    from timeit import default_timer as timer

    start = timer()
    match_by_max_overlap(
        '/E/Sample_images/sample_images_PA/trash_test_mem/trash_registration_network/1/210930_EcadKI_mel_40-54hAPF_ON.lif - Series011_t001.tif',
        '/E/Sample_images/sample_images_PA/trash_test_mem/trash_registration_network/1/210930_EcadKI_mel_40-54hAPF_ON.lif - Series011_t000_fake.tif')
    # match_by_max_overlap(Img('/E/Sample_images/sample_images_PA/trash_test_mem/trash_registration_network/1/210930_EcadKI_mel_40-54hAPF_ON.lif - Series011_t001/handCorrection.tif'), RGB_to_int24(Img('/E/Sample_images/sample_images_PA/trash_test_mem/trash_registration_network/1/warped_mask.tif')))
    # match_by_amx_overlap(Img('/E/Sample_images/sample_images_PA/trash_test_mem/trash_registration_network/1/210930_EcadKI_mel_40-54hAPF_ON.lif - Series011_t001/handCorrection.tif'),RGB_to_int24(Img('/E/Sample_images/sample_images_PA/trash_test_mem/trash_registration_network/1/210930_EcadKI_mel_40-54hAPF_ON.lif - Series011_t000/tracked_cells_resized.tif'))) # ça marche vraiment mieux en utilisant la pre registration en fait!!!

    # need pre reg it




    print('stop', timer() - start)

    # faire une pipeline qui peut etre va generer un shifted track or un shifted --> ideally I should just put this code within mermaid --> it would be a new tracking --> that would read the cell label file and

    # finalize this



# try to find a code that finds the new scaling so that the image fits to the new size
# assume all images have the same height in a row and the same width in a column

# can I not even store everything as a big numpy array ???

# test with 3 rectangles

# keep AR the same --> need just apply a scaling to the image and in fact to all images the same, by the way

#
# def find_scaling_to_fit():
#     incompressible_length = (len(widths)-1)*space_between_images
#     image_width_to_fit_not_counting_incompressible_length = final_width-incompressible_length
#     sum_individual = sum(widths)
#     common_scaling = image_width_to_fit_not_counting_incompressible_length/sum_individual
#
#     print(common_scaling)
#
#     # print(widths*common_scaling)
#     corrected_widths = [w*common_scaling for w in widths]
#     print(widths)
#     print(corrected_widths)
#     print(sum(corrected_widths))
#     print(sum(corrected_widths)+incompressible_length)
#     print(incompressible_length)
#     # print()

    # ça marche super en fait et c'est vraiment facile

# computes the rescaling factor so that images get rescaled to fit the desired dimension --> much smarter that way than using brute force!!!

import os
from epyseg.settings.global_settings import set_UI # set the UI to be used py qtpy
set_UI()
import sys
from qtpy import QtWidgets
from qtpy.QtCore import QRectF


# nb the pb asssumes AR is kept constant which is true for images but not with panels containing images and incompressible space between them --> need some extra computation then


common_value = 0


def get_common_scaling_factor(dimensions_in_the_desired_dimension, desired_final_size, size_of_incompressible_elements_in_dimension=3):
    incompressible_length = (len(dimensions_in_the_desired_dimension) - 1) * size_of_incompressible_elements_in_dimension
    image_width_to_fit_not_counting_incompressible_length = desired_final_size - incompressible_length
    sum_individual = sum(dimensions_in_the_desired_dimension)
    common_scaling = image_width_to_fit_not_counting_incompressible_length / sum_individual

    # print(common_scaling)

    # print(widths*common_scaling)
    # corrected_widths = [w * common_scaling for w in dimensions_in_the_desired_dimension]
    # print(dimensions_in_the_desired_dimension)
    # print(corrected_widths)
    # print(sum(corrected_widths))
    # print(sum(corrected_widths) + incompressible_length)
    # print(incompressible_length)

    return common_scaling

#########################################################KEEP EQUATIONS TO SOLVE THE CHANGE IN SIZE
# final_width = width_obj1 + width_obj2+ width_obj3 + (nb_of_objs-1)*space
# new_width = new_width_obj1 + new_width_obj2+ new_width_obj3 + (nb_of_objs-1)*space
# new_width = width_obj1*scaling + new_width_obj2*scaling+ new_width_obj3*scaling + (nb_of_objs-1)*space
# height = same whatever the condition but can change --> need solve the equa so that height is the same while width obeys the new stuff
# heightA = heightB = heightC # always
# comment relier la width à la height ???

# final_AR = widthA/heightA = (incomp_widthA+widthA_non_incom)/(incomp_heightA+heightA_non_incom)
# heightA =(incomp_heightA+heightA_non_incom)
# widthA = (incomp_widthA+withA_non_incom)
# AR = withA_non_incom/heightA_non_incom
# heightA_non_incomp = widthA_non_incomp/AR
# heightA = incomp_heightA+(widthA_non_incomp/AR)
# widthA = (incomp_widthA+withA_non_incom)
# --> with that can I compute the best change for all
# final_AR = ((scaling_x*current_widthA) +widthA_non_incom)/((scaling_y*current_heightA)+heightA_non_incom)
# new_widthA = ((scaling_xA*current_widthA) +widthA_non_incom)
# new_heightA = ((scaling_yA*current_heightA)+heightA_non_incom)
# I need find both scalings at the same time
# reorder equa for scaling -> TODO
# scaling_xA = (new_widthA - widthA_non_incom)/current_widthA
# scaling_yA = (current_heightA - heightA_non_incom)/current_heightA
# need find both so that everything fits
# --> easy in fact and that is what I need to have
# height is always related to width too given the incompressible AR ratio --> YES IN FACT IT MUST BE
# heightA = heightB = heightC # whatever happens the images all have the same height in a row
# so by definition
# height A is fixed but yet to an unknown value
# heightA = heightB = heightC
# width is fixed and height must be adjusted but the same in all cases
# --> how does one compute that
# that must be doable but complex
# j'ai 6 inconnues --> mes scalings en x et en y et j'ai une taille constante pour un et une egalite pour l'autre
#


# final AR = ((scalingx * old_compressible_dim_x)+incomp x)/((scalingy * old_compressible_dim_y)+incomp y)) =  ((scalingx * old_compressible_dim_x)+incomp x)/((scalingy * (old_compressible_dim_x/AR))+incomp y))
# finalAR = final_dimx/final dimy
# final_dimx/final dimy = ((scalingx * old_compressible_dim_x)+incomp x)/((scalingy * (old_compressible_dim_x/AR))+incomp y))
# final_dimx = final dimy* (((scalingx * old_compressible_dim_x)+incomp x)/((scalingy * (old_compressible_dim_x/AR))+incomp y)))
# final_dimy = final dimx/(((scalingx * old_compressible_dim_x)+incomp x)/((scalingy * (old_compressible_dim_x/AR))+incomp y)))
# final_dimy1 = final_dimy2 =final_dimy3
# the early things are known and either the final height or width is known --> not true but i know they are eaqual --> so I can solve it
# and get all the parameters

# sum of dims + sum incomp = desired width
# sum(incomp)+sum(images) = desired_width
# I need get the 3 heights that would fulfill the criterion I want


# same height --> ????
# how do I do that






# faire ça

# start with just two images and solve it
# heightA = heightB
# final_width = widthA + widthB + incomp_width
# need find widthA and widthB so that the heights are the same
# final_width = scalingxA * original_widthA +scalingxB * original_widthB  + incomp_width
# final_height = scalingyA * original_heightA+scalingyB * original_heightB
# how can I relate width and height
# final_AR is unknown

# in fact I don't find any other way than brute force but there has to be a way ??? --> too bad I forgot my basic math classes
# ratio = sizex/sizey --> ce n'est pas fixe mais ça depend des
# the pb is that this ratio is unknown


# height +  ((leny-1)*space) = AR*width + ((lenx-1)*space)
# height = AR*width + ((lenx-1)*space)-  ((leny-1)*space)
# heights of the three should be equal --> can compute
# y = ax+b # image1 --> what I want is a ???
# y = dx'+c
# all same height
#dx'+c = ax+b
# sum of width = fixed size
# dx'+c + ax+b= 512
# c and b are fixed
# dx'+ax = 512-c-b # --> equation 1
# need another equation
# need find d and b
# est ce que c'est ça ???
# https://calculis.net/systeme-n-equations#solutions

# dx'+c = ax+b
# d = (ax+b-c)/x'
# a = (dx'+c-b)/x
# deux equations à deux inconnues
# https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.fsolve.html or that ???
# https://numpy.org/doc/stable/reference/generated/numpy.linalg.solve.html # or that

# or solve for different heights
# y1 = a1*x1+b1
# y2 = a2*x2+b2
# y1 = y2
# a1x1 = a2*x2+b1-b2
# a1*x1+b1 + a2*x2+b2 + c = 512
# https://www.youtube.com/watch?v=QkkpcVbNLVE --> je pense que c'est ce que je cherche
# du coup voir comment le resoudre pratiquement
# or brute force solve it numerically
# see the smart and fastest way to do that
# i need find a1 and x1 and a2 and x2 -> need another equation then



# autre solution faire usr un systeme packé et le changer en non packé --> marche si et seulement si le truc est incompressible dans une seule dimension
#


















#########################################################KEEP EQUATIONS TO SOLVE THE CHANGE IN SIZE

# worst case --> I can brute force it a bit but I would love not to
# or need a post process correction to fit in same height but pb is that it will also change width and therefore I will have trouble beacuse it will change width and I will need do that recursively until I reach the global minima
# -> think about it
# think about it

# AR = (real_width + incomp_width)/(real_height + incomp_height)
# incomp_width =
# width = sum_width_objects
# probably easy --> I just need to have in the formula the AR
# it should fit in the given space
# if same size --> incompressible fits
def get_common_scaling_factor_taking_incompressibility_into_account(dimensions_in_the_desired_dimension, desired_final_size, size_of_incompressible_elements_in_dimension=3):
    incompressible_length = (len(dimensions_in_the_desired_dimension) - 1) * size_of_incompressible_elements_in_dimension
    image_width_to_fit_not_counting_incompressible_length = desired_final_size - incompressible_length
    sum_individual = sum(dimensions_in_the_desired_dimension)
    common_scaling = image_width_to_fit_not_counting_incompressible_length / sum_individual

    # print(common_scaling)

    # print(widths*common_scaling)
    # corrected_widths = [w * common_scaling for w in dimensions_in_the_desired_dimension]
    # print(dimensions_in_the_desired_dimension)
    # print(corrected_widths)
    # print(sum(corrected_widths))
    # print(sum(corrected_widths) + incompressible_length)
    # print(incompressible_length)

    return common_scaling

def get_master_bounds2(group_of_shapes):
    bounds = QRectF()
    max_width = 0
    max_height = 0
    min_x=100000000
    min_y=100000000
    for shape in group_of_shapes:
       rect = shape.boundingRect()
       max_width = max(max_width, rect.x()+rect.width())
       max_height = max(max_height, rect.y()+rect.height())
       min_x = min(min_x, rect.x())
       min_y = min(min_y, rect.y())
    bounds.setX(min_x)
    bounds.setY(min_y)
    bounds.setWidth(max_width-min_x)
    bounds.setHeight(max_height-min_y)
    return bounds
#
# gets the master rect of a list of objects
# Nb should I allow negative bounds ??? --> maybe not in fact
def get_master_bounds(group_of_shapes):
    bounds = QRectF()
    max_width = 0
    max_height = 0
    for shape in group_of_shapes:
       rect = shape.boundingRect()
       max_width = max(max_width, rect.x()+rect.width())
       max_height = max(max_height, rect.y()+rect.height())
    bounds.setWidth(max_width)
    bounds.setHeight(max_height)
    return bounds

def preview( shapes_to_draw):
    if not isinstance(shapes_to_draw, list):
        shapes_to_draw=[shapes_to_draw]
    from epyseg.figure.ezfig import MyWidget
    import sys
    app = QtWidgets.QApplication(sys.argv)
    # app = QGuiApplication(sys.argv)
    widget = MyWidget()

    widget.shapes_to_draw = shapes_to_draw
    widget.update_size()

    widget.show()

    sys.exit(app.exec_())


if __name__ == '__main__':


    if True:
        # test solving equation
        import numpy as np

        # images have same height --> can fix that to the desired value
        # y = a0*x0+b0
        # y = a1*x1+b1
        # y-b0 = a0*x0
        # y-b1 = a1*x1

        # image 1 128*256
        # image 2 128*128
        # image 3 64*128
        # 256 = a0 * 128 --> 128 = a0 * 64 # image1
        # 128 = a1 * 128 -> im2
        # 64 = a2* 128

        # ça ne marche pas
        # manque une equa qui me donne ce que je veux
        a = np.array([[64, 0,0], [0, 128,0],[0, 0,128]]) #--> indeed need scale it by a factor2 --> # why not 0.5
        b = np.array([128,128,64])
        x = np.linalg.solve(a, b)
        print(x)


        # somme des trois -->
        # a = np.array([[64, 0, 0], [0, 128, 0], [0, 0, 128]])  # --> indeed need scale it by a factor2 --> # why not 0.5
        # b = np.array([128, 128, 64])
        # x = np.linalg.solve(a, b)
        # print(x)
        # j'ai n equations --> comment ensuite les combiner

        # sinon calculer les images à trois hauteurs differentes et calculer comment obtenir le truc à la hauteur souhaitée --> à faire
        # facile de faire des set to height et de mesurer les valeurs
        # a = np.array([[512, 0,0], [0, 512,0], [512,512,1024]])  # --> indeed need scale it by a factor2 --> # why not 0.5
        # b = np.array([512, 512, 1024])
        # x = np.linalg.solve(a, b)
        # print(x)

        a = np.array([[64, 0], [64, 128]]) #--> indeed need scale it by a factor2 --> # why not 0.5
        b = np.array([128,256])
        # 2 et 1 -->
        # 256 = 2*128+128 # --> ok but not what I want
        x = np.linalg.solve(a, b)
        print(x)

        # n = len(ni)
        # from sympy import symbols, solve
        #
        # A = symbols('a0:%d' % n)
        # equations = []  # set up equations list
        # for i in range(n):
        #     newsum = (rSv + (A[i] + ni[1]) * nb[i])
        #     equations.append(Eq((A[i] + d2sni[i]) * nb[i] / newsum, S[i]))  # append instead of overwriting equations.
        # sol = solve(equations)

        # is that what I want
        # from scipy.optimize import fsolve
        # import math
        #
        #
        # def equations(p):
        #     x, y = p
        #     return (x + y ** 2 - 4, math.exp(x) + x * y - 3)
        #
        #
        # x, y = fsolve(equations, (1, 1))
        #
        # print(equations((x, y)))



        # https://docs.sympy.org/dev/modules/solvers/solvers.html#sympy.solvers.solvers.nsolve
        # import sympy
        from sympy import nsolve, exp, Symbol
        # The first argument is a list of equations, the second is list of variables and the third is an initial guess.
        y = Symbol('y')
        x= Symbol('x')
        # y = a*x
        # y-ax
        # comment faire ????
        print(nsolve([x + y ** 2 - 4, exp(x) + x * y - 3], [x, y], [1, 1]))
        # --> ok that works
        # can I formulate my math this way

        # final_dimx = final dimy* (((scalingx * old_compressible_dim_x)+incomp x)/((scalingy * (old_compressible_dim_x/AR))+incomp y)))
        # final_dimy = final dimx/(((scalingx * old_compressible_dim_x)+incomp x)/((scalingy * (old_compressible_dim_x/AR))+incomp y)))
        # final_dimy1 = final_dimy2 =final_dimy3
        # sum of x = final_dimx+final dimx2+final dimx3 + sum(incomp)


        # images 512*512 AR 1
        # image 128*512 AR 0.25
        # image 256*256 AR 1
        # incomp width = 3*2
        # desired width = 512
        a1 = Symbol('a1')
        a2 = Symbol('a2')
        a3 = Symbol('a3')


        # cette version marche mais faut voir
        print('test2',nsolve([a1*512+a2*128-(512-3), a1*1*512+0-a2*(1/0.25)*256-0], [a1, a2], [1, 1]))
        # print('test3',nsolve([a1*512+a2*128+a3*256-(512-6), (a1*1*512-0)+(a2*(1/0.25)*256-0)-2*(a3*1*256-0), 2*(a1*1*512-0)-(a2*(1/0.25)*256-0)-(a3*1*256-0)], [a1, a2,a3], [1, 1,1]))
        # print('test4',nsolve([(a1*512+a2*128+a3*256)-(512-6), (a1*1*512-0)-(a2*(1/0.25)*256-0), (a2*(1/0.25)*256-0)-(a3*1*256-0)], [a1, a2,a3], [1, 1,1])) # ça a l'air de marche en fait --> à voir
        print('test4',nsolve([a1*512+a2*128+a3*256+6-512, (a1*1*512+0)-(a2*(1/0.25)*128+0), (a2*(1/0.25)*128-0)-(a3*1*256-0)], [a1, a2,a3], [1, 1,1])) # ça a l'air de marche en fait --> à voir
        # print('test4',nsolve([a1*512+a2*128+a3*256+6-512, (a1*1*512+0)-(a2*0.25*128+0), (a2*0.25*128-0)-(a3*1*256-0)], [a1, a2,a3], [1, 1,1])) # ça a l'air de marche en fait --> à voir
        # 235.52 + 29.764705882352896 +238.117647058823424 # --> marche pas du tout en fait
        # --> un bug ???
        # 224.768 + 224.768 + 224.888888888888832 # colle pas mais la hauteur est en effet la meme --> mais pb pas bonne largeur --> bug dans mon equa ???

        # test4 Matrix([[0.0470610119047619], [0.752976190476190], [1.50595238095238]])
        print('test5',nsolve([a1*512+a2*128+a3*256+6-512, (a1*1*512+0)-(a2*(1/0.25)*128+0), (a2*(1/0.25)*128-0)-(a3*1*256-0)], [a1, a2,a3], [1, 2,2])) # ça a l'air de marche en fait --> à voir
        test = nsolve([a1*512+a2*128+a3*256+6-512, (a1*1*512+0)-(a2*(1/0.25)*128+0), (a2*(1/0.25)*128-0)-(a3*1*256-0)], [a1, a2,a3], [1, 2,2])
        dims_x = [512,128,256]
        dims_y = [512,512,256]
        summ_x = []
        summ_y = []
        for iii,val in enumerate(test):
            print('-->', dims_x[iii]*val, val)
            summ_x.append(dims_x[iii]*val)
            summ_y.append(dims_y[iii]*val)

        print('sumX', sum(summ_x)+6) # --> si ça marche en fait
        print('sumY', summ_y) # --> si ça marche en fait

        # --> Matrix([[0.883680555555556], [0.441840277777778]])

        # 153.6
        # 339.2
        # --> 492
        # --> a l'air ok --> voir sure un truc plus complexe
        # voir aussi comment générer les equas de maniere 100% autonome
        # --> pas trop dur je pense...

        # all seems ok --> just try with a more complex shape to see if that works
        # see also how to handle groups --> is that easy or do they need be decomposed into smaller things ???

        if True:
            from epyseg.draw.shapes.image2d import Image2D
            from epyseg.figure.column import Column
            from epyseg.figure.row import Row
            # from epyseg.draw.shapes.image2d import Image2D
            img1 = Image2D('/E/Sample_images/sample_images_PA/trash_test_mem/counter/00.png')
            img2 = Image2D('/E/Sample_images/sample_images_PA/trash_test_mem/counter/01.png')
            img3 = Image2D('/E/Sample_images/sample_images_PA/trash_test_mem/counter/02.png')
            img4 = Image2D('/E/Sample_images/sample_images_PA/trash_test_mem/counter/03.png')
            img5 = Image2D('/E/Sample_images/sample_images_PA/trash_test_mem/counter/04.png')
            img6 = Image2D('/E/Sample_images/sample_images_PA/trash_test_mem/counter/05.png')
            img7 = Image2D('/E/Sample_images/sample_images_PA/trash_test_mem/counter/06.png')
            img8 = Image2D('/E/Sample_images/sample_images_PA/trash_test_mem/counter/07.png')
            img9 = Image2D('/E/Sample_images/sample_images_PA/trash_test_mem/counter/08.png')


            images = [img1, img2, img3, img4]

            # get AR and stuff for all the images --> TODO
            # solution (128.00066196944888, 42.80999999999984, 42.80999999999984)
            space = 3
            col1 = Column(img1, img2, space=space)
            # col2 = Column(img3, img4, space=space)
            # col1 /= col2

            row1 = Row(img5, img6,  space=space) # img7,

            col1 |= row1
            # there is a bug in there in fact --> it starts not to work with very complex objects
            # --> try it and do my best
            # somehow need same height and also stuff

            # col1.setToHeight(256) # --> ça marche par contre la set
            # col1.setToWidth(256) # --> ça marche par contre la set to width ne marche pas --> probably needs be fixed because the user will want that

            # col1.setToWidth(128)

            # get width for all and incomp for all

            # (128.01439372226696, 40.469999999999914, 40.46999999999991) --> avec juste 4 images --> voir la vraie solution
            print(img1.width(False))
            print(img2.width(False))
            print(img5.width(False))
            print(img6.width(False))
            # print(img7.width(False))

            print('h',img1.height(False))
            print(img2.height(False))
            print(img5.height(False))
            print(img6.height(False))
            # print(img7.height(False))

            print('AR', img1.width(False)/img1.height(False))
            print(img2.width(False)/img2.height(False))
            print(img5.width(False)/img5.height(False))
            print(img6.width(False)/img6.height(False))
            # print(img7.width(False)/img7.height(False))



            # incomp height = 3 pr col et 0 pr autres
            # incomp_w=6

            # test = nsolve([a1 * 512 + a2 * 128 + a3 * 256 + 6 - 512, (a1 * 1 * 512 + 0) - (a2 * (1 / 0.25) * 128 + 0),(a2 * (1 / 0.25) * 128 - 0) - (a3 * 1 * 256 - 0)], [a1, a2, a3], [1, 1, 1])
            # need get equa per stuff
            # --> see how to do that

            # get equa for row
            # if nested equa --> TODO --> return nested equa
            # height

            # 288.0
            # 145.0
            # 299.0
            # 299.0
            # h
            # 322.0
            # 303.0
            # 352.0
            # 299.0
            # AR
            # 0.8944099378881988
            # 0.47854785478547857
            # 0.8494318181818182
            # 1.0

            a1 = Symbol('a1')
            a2 = Symbol('a2')
            a3 = Symbol('a3')
            a4 = Symbol('a4')

            # for each row in stuff --> do for all elements of the row


            # test = nsolve([a1 * 512 + a2 * 128 + a3 * 256 + 6 - 512, (a1 * 1 * 512 + 0) - (a2 * (1 / 0.25) * 128 + 0),(a2 * (1 / 0.25) * 128 - 0) - (a3 * 1 * 256 - 0)], [a1, a2, a3], [1, 1, 1])

            # need solve for width and for height --> need different equas for rows and columns


            # test = nsolve([a1 * 288 + a2 * 145 + a3 * 299 + 6 - 512, (a1 * (1/0.8944099378881988) * 512 + 0) - (a2 * (1 / 0.47854785478547857) * 145 + 0),(a2 *(1 / 0.47854785478547857) * 145 - 0) - (a3 * (1/0.8494318181818182) * 299 - 0)], [a1, a2, a3], [1, 1, 1])
            # dims_x = [288,145, 299]
            # dims_y = [322, 303, 352]
            # summ_x = []
            # summ_y = []
            # for iii, val in enumerate(test):
            #     print('-->', dims_x[iii] * val, val)
            #     summ_x.append(dims_x[iii] * val)
            #     summ_y.append(dims_y[iii] * val)
            #
            # print('sumX', sum(summ_x) + 6)  # --> si ça marche en fait
            # print('sumY', summ_y)  # --> si ça marche en fait

            #ça marche mais si une equa est complexe alors il faut mettre l'equa complexe dedans avec ses variables aussi
            # --> voir comment faire --> faut faire generer une equa par truc ...
            # --> voir comment faire et faire generer les equas par row et par col --> des sous equas du coup

            # peut etre résoudre la meme pour chaque colonne en fait


            # for a ROW
            dims_x = []
            dims_y = []
            summ_x = []
            summ_y = []
            equations_single = []
            expressions_a = []
            expressions_b = []
            expressions_c = []
            equations = []
            variables = []
            estimates = []
            for iii, img in enumerate(images):
                dims_x.append(img.width(False))
                dims_y.append(img.height(False))
                AR = img.width(False)/img.height(False)

                variable = Symbol('a'+str(iii))
                variables.append(variable)
                estimates.append(1)
                equations_single.append(variable*(1./AR)*img.width(False)+0)
                expr = variable * (1. / AR) * img.width(False) + 0
                expressions_a.append(expr)
                expr= variable *img.width(False)
                expressions_b.append(expr)

                # expr = variable*AR*img.height(False) # this is the new width
                # expr = variable*AR*img.height(False) # this is the new width
                #

            # add an equa that is the sum of all
            # final_equa = # sum of all and substraction
            # final_equa = []

            print(estimates, equations_single, variables, expressions_b)
            final_equa = None
            for expr in expressions_b:
                if final_equa is None:
                    final_equa = expr
                else:
                    final_equa+=expr



            # --> equas are all heights should be same

            desired_width = 512
            final_equa+= (len(images)-1)*space -desired_width
            print(final_equa)  # --> ok that works

            # do by pairs of equal
            pairs_of_equations = None
            other_equas = []
            for iii in range(1,len(expressions_a)):
                # print(iii)
                # first - sec = 0
                equa =  expressions_a[0]-expressions_a[iii]
                other_equas.append(equa)

            other_equas.append(final_equa)

            print('toto',other_equas, variables, len(expressions_a))

            test = nsolve(other_equas,variables,estimates)
            summ_x = []
            summ_y = []
            for iii, val in enumerate(test):
                print('-->', dims_x[iii] * val, val)
                summ_x.append(dims_x[iii] * val)
                summ_y.append(dims_y[iii] * val)

            print('sumX', sum(summ_x) +  (len(images)-1)*space)  # --> si ça marche en fait
            print('sumY', summ_y)  # --> si ça marche en fait

            # --> cool ça marche
            # -->
            # is that all ok
            # for a COL
            # dims_x = []
            # dims_y = []
            # summ_x = []
            # summ_y = []
            # equations_single = []
            # expressions_a = []
            # expressions_b = []
            # equations = []
            # variables = []
            # estimates = []
            # for iii, img in enumerate(images):
            #     dims_x.append(img.width(False))
            #     dims_y.append(img.height(False))
            #     AR = img.width(False) / img.height(False)
            #
            #     variable = Symbol('a' + str(iii))
            #     variables.append(variable)
            #     estimates.append(1)
            #     equations_single.append(variable * (1. / AR) * img.width(False) + 0)
            #     expr = variable * (1. / AR) * img.width(False) + 0
            #     expressions_a.append(expr)
            #
            #     expr = variable * img.width(False)
            #     expressions_b.append(expr)
            #
            # # add an equa that is the sum of all
            # # final_equa = # sum of all and substraction
            # # final_equa = []
            #
            # print(estimates, equations_single, variables, expressions_b)
            # final_equa = None
            # for expr in expressions_b:
            #     if final_equa is None:
            #         final_equa = expr
            #     else:
            #         final_equa += expr
            #
            # desired_width = 512
            # final_equa += (len(images) - 1) * space - desired_width
            # print(final_equa)  # --> ok that works
            #
            # # do by pairs of equal
            # pairs_of_equations = None
            # other_equas = []
            # for iii in range(1, len(expressions_a)):
            #     # print(iii)
            #     # first - sec = 0
            #     equa = expressions_a[0] - expressions_a[iii]
            #     other_equas.append(equa)
            #
            # other_equas.append(final_equa)
            #
            # print('toto', other_equas, variables, len(expressions_a))
            #
            # test = nsolve(other_equas, variables, estimates)
            # summ_x = []
            # summ_y = []
            # for iii, val in enumerate(test):
            #     print('-->', dims_x[iii] * val, val)
            #     summ_x.append(dims_x[iii] * val)
            #     summ_y.append(dims_y[iii] * val)
            #
            # print('sumX', sum(summ_x) + (len(images) - 1) * space)  # --> si ça marche en fait
            # print('sumY', summ_y)  # --> si ça marche en fait






            # test autosolver for a row at least --> TODO


            # remettre mes nouvelles equas ici:



        # panel 1 ---> imageA | imageB --> incompx = 3 incompy = 0
        # panels2 imageC/imageD/imageE -- >  incompx = 0 incmp y =6
        # final fig = panel1 | panels2
        # all is ok for now
        #

        # do it for compressed chape then compute the 3 changes of AR that are needed to incorporate within the same fig the incompressible part of the image --> maybe three equations so I can solve them
        # --> TODO
        # worst case scenario --> back to brute force...

        # imageA -E --> 256w * 128h
        #
        # packed width --> 128 | 256|256|256 --> width = 128+256*3 -> 896
        # packed height --> 128 | 128|128|128 --> height = 128
        # if I wanna add to it spacer I need to do -->
        # if I wanna add a space between them --> I need add 3 pixels
        #  128 |3| 256|256|256 --> new width 896+3 --> 899--> if I want them to fit in 896 I need scale both stuff that are fully compressible by a value that is
        # I need * width and height by 896/897 =0.99888517279821627648 then it will fit --> easy because compressible and that would also change the height










        # test 3 width
        # x1 + x2 + x3 + sum incompressible = desired_width
        # rescaling
        # y1+y2 = 2*y3 --> mes equations
        # en fait y peut aussi etre part de mon equation




        # all seems ok --> can i now solve all my equations
        # from sympy import Symbol, nsolve
        # import mpmath
        # mpmath.mp.dps = 15
        # x1 = Symbol('x1')
        # x2 = Symbol('x2')
        # f1 = 3 * x1 ** 2 - 2 * x2 ** 2 - 1
        # f2 = x1 ** 2 - 2 * x1 + x2 ** 2 + 2 * x2 - 8
        # print(nsolve((f1, f2), (x1, x2), (-1, 1)))










        # 3 images 512*512
        # on veut à la fin 1536 --> scaling factor =1 pr chaque
        # finaliser aussi le code des clones
        # --> faire ça avec et ou sans divisions --> à faire donc
        #
        # put the n equations
        # a1*




        # 128 = 2*64
        # 128 = 1* 128
        # mais du coup le facteur de retaillage c'est pas ça
        # essayer


        # poser les eaqution pour que la hauteur soit la meme

        # how can I get my n equations
        # Solve the system of equations x0 + 2 * x1 = 1 and 3 * x0 + 5 * x1 = 2:
        # array([-1.,  1.]) # --> -1+2*1=1 3*-1+5*1=2 --> ça marche
        # that seems to work --> image same height -> 2 et 1
        # a0 = 2 et a1 = 1
        # facteur 2
        # a1x1+a2x2+a3x3 = 512-b1-b2-b3
        # une equation


        # desired width = 128
        # real example from columns --> with space 3
        # size of inner elm qtpy.QtCore.QRectF(0.0, 0.0, 6.491758918437604, 49.188463457232196)
        # size of inner elm qtpy.QtCore.QRectF(9.491758918437604, 0.0, 118.50824108156237, 42.42126355836911)
        # <class 'epyseg.figure.row.Row'>
        # qtpy.QtCore.QRectF(0.0, 0.0, 127.99999999999997, 49.188463457232196)

        # with space 0
        #size of inner elm qtpy.QtCore.QRectF(0.0, 0.0, 7.348408669130177, 45.49171603236653)
        # size of inner elm qtpy.QtCore.QRectF(7.348408669130176, 0.0, 120.65159133086982, 45.49171603236654)
        # <class 'epyseg.figure.row.Row'>
        # qtpy.QtCore.QRectF(0.0, 0.0, 128.0, 45.49171603236654)
        # if I put them to sameheight --> then with size 3 --> qtpy.QtCore.QRectF(0.0, 0.0, 145.9477387942681, 49.188463457232196) --> not 128 anymore --> need resize it
        # need change the height from there one by one so that I reach 128 in width again


        # a = np.array([[1, 2], [3, 5]])
        # b = np.array([1, 2])
        # x = np.linalg.solve(a, b)
        # print(x)



        sys.exit(0)

    # find_scaling_to_fit()

    widths = [30, 15, 12, 3]
    final_width = 120
    space_between_images = 3  # 3px

    print(get_common_scaling_factor(widths, final_width, space_between_images))
    # so cool --> that will always work --> I love it
    # tester en grandeur reelle


    # ça marche --> can even preview a single image --> good
    from epyseg.draw.shapes.image2d import Image2D
    img = Image2D('/E/Sample_images/sample_images_EZF/counter/00.png')
    preview(img)




    # faire un autopanel maybe ???
    # or by default do a figure
    # try it all and see what I can do also add plots --> figures --> either from pandas or from a figure from matplotlib --> TODO






import os
from epyseg.settings.global_settings import set_UI # set the UI to be used py qtpy
set_UI()
from qtpy import QtCore
from qtpy.QtCore import QPointF, Qt
from qtpy.QtGui import QPolygonF, QTransform
from qtpy.QtGui import QPainter, QBrush, QPen, QImage, QColor
# from qtpy.Qt
# from qtpy.Qt import (QPaintEngine, QPaintDevice,  QTransform, QBrush)

from epyseg.tools.logger import TA_logger
logger = TA_logger()

class Polygon2D(QPolygonF):


    def __init__(self, *args, color=0xFFFF00, fill_color=None, opacity=1., stroke=0.65, line_style=None, theta=0, **kwargs):
        super(Polygon2D, self).__init__()
        self.isSet = False
        # print(args, len(args))
        if len(args) > 0:
            for i in range(0, len(args), 2):
                self.append(QPointF(args[i], args[i+1]))
            self.isSet = True
                # print(QPointF(args[i], args[i+1]))

        self.color = color
        self.fill_color = fill_color
        self.stroke = stroke
        self.opacity = opacity
        self.scale = 1
        self.translation = QPointF()
        self.line_style = line_style
        # self.isSet = False
        # rotation
        self.theta = theta

    def set_rotation(self, theta):
        self.theta = theta

    def set_opacity(self, opacity):
        self.opacity = opacity

    def set_line_style(self,style):
        """
        allows lines to be dashed or dotted or have custom pattern

        :param style: a list of numbers or any of the following Qt.SolidLine, Qt.DashLine, Qt.DashDotLine, Qt.DotLine, Qt.DashDotDotLine but not Qt.CustomDashLine, Qt.CustomDashLine is assumed by default if a list is passed in. None is also a valid value that resets the line --> assume plain line
        :return:
        """
        self.line_style = style
        # if style is a list then assume custom pattern otherwise apply solidline

    def get_color(self):
        return self.color

    def get_fill_color(self):
        return self.fill_color

    def get_stroke_size(self):
        return self.stroke

    def get_points(self):
        points = []
        for point in self:
            points.append((point.x(), point.y()))
        return points

    def contains(self, *args):
        return self.containsPoint(*args, 0)

    def draw(self, painter, draw=True):
        if self.color is None and self.fill_color is None:
            return

        if draw:
            painter.save()
            painter.setOpacity(self.opacity)
            if self.color is not None:
                pen = QPen(QColor(self.color))
                if self.stroke is not None:
                    pen.setWidthF(self.stroke)
                if self.line_style is not None:
                    if self.line_style in [Qt.SolidLine, Qt.DashLine, Qt.DashDotLine, Qt.DotLine, Qt.DashDotDotLine]:
                        pen.setStyle(self.line_style)
                    elif isinstance(self.line_style, list):
                        pen.setStyle(Qt.CustomDashLine)
                        pen.setDashPattern(self.line_style)
                painter.setPen(pen)
            else:
                painter.setPen(Qt.NoPen)  # required to draw something filled without a border
            if self.fill_color is not None:
                painter.setBrush(QBrush(QColor(self.fill_color)))
            else:
                painter.setBrush(Qt.NoBrush)
            polygon_to_draw = self.translated(0, 0)
            if self.scale is not None and self.scale != 1:
                polygon_to_draw = self.__scaled()

            # print('mid rect_to_plot', rect_to_plot)
            if self.translation is not None:
                # rect_to_plot.setX(rect_to_plot.x()+self.translation.x())
                # rect_to_plot.setY(rect_to_plot.y()+self.translation.y())
                polygon_to_draw.translate(self.translation.x(), self.translation.y())

            if self.theta is not None and self.theta != 0:
                painter.translate(polygon_to_draw.center())
                painter.rotate(self.theta)
                painter.translate(-polygon_to_draw.center())

            painter.drawPolygon(polygon_to_draw)
            painter.restore()

    # def fill(self, painter, draw=True):
    #     if self.fill_color is None:
    #         return
    #     if draw:
    #         painter.save()
    #     painter.setBrush(QBrush(QColor(self.fill_color)))
    #     painter.setOpacity(self.opacity)
    #     if draw:
    #         painter.drawPolygon(self)
    #         painter.restore()
    #
    # def drawAndFill(self, painter):
    #     painter.save()
    #     self.draw(painter, draw=False)
    #     self.fill(painter, draw=False)
    #     painter.drawPolygon(self)
    #     painter.restore()

    # def setP1(self, point):
    #     self.append(point)
    def get_P1(self):
        return self.boundingRect().topLeft()

    def set_P1(self, point):# see how to do that
        #self.append(point)
        # get bounding box and set top left to P1
        current_pos = self.boundingRect().topLeft()
        self.translate(point.x()-current_pos.x(), point.y()-current_pos.y())

    def add(self, *args, force=False):
        if self.count() > 1 and not force:
            self.remove(self.count()-1)
        self.append(args[1])
        self.isSet = True

    def listVertices(self):
        return [point for point in self]

    def set_to_scale(self, factor):
        self.scale = factor

    def set_to_translation(self, translation):
        self.translation = translation

    def __scaled(self):
        vertices = self.listVertices()
        scaled_poly = QPolygonF()
        for vx in vertices:
            vx.setX(vx.x()*self.scale)
            vx.setY(vx.y()*self.scale)
            scaled_poly.append(vx)
        return scaled_poly

    def boundingRect(self):
        polygon_to_draw = self.translated(0, 0)
        try:
            # print('tada')
            if self.theta is not None and self.theta != 0:
                # print('entering')
                center = polygon_to_draw.boundingRect().center()
                # print('entering2')
                t = QTransform().translate(center.x(), center.y()).rotate(self.theta).translate(-center.x(),
                                                                                                -center.y())
                #  print('entering3')
                #  transformed = self.setTransform(t)
                #  print('entering4')
                #  print(transformed)
                #  print(QRectF(min(transformed.p1().x(), transformed.p2().x()), min(transformed.p1().y(), transformed.p2().y()),
                # abs(transformed.p2().x() - transformed.p1().x()), abs(transformed.p2().y() - transformed.p1().y())))
                #  return QRectF(min(transformed.p1().x(), transformed.p2().x()), min(transformed.p1().y(), transformed.p2().y()),
                # abs(transformed.p2().x() - transformed.p1().x()), abs(transformed.p2().y() - transformed.p1().y()))

                # copy.setT
                # print('entering')

                # t = QTransform().translate( center.x(), center.y()).rotate(self.theta).translate(-center.x(), -center.y())
                # # print('entersd')
                transformed = t.map(polygon_to_draw)  # // mapRect() returns the bounding rect of the rotated rect

                # print('rotated',rotatedRect )
                # return rotatedRect
                return transformed.boundingRect()
        except:
            pass
        return polygon_to_draw.boundingRect()

if __name__ == '__main__':

    test = Polygon2D(0, 0, 10, 0, 10, 20, 0, 20, 0, 0)
    print(test.count()) # marche pas --> pas ajouté
    print(test)

    hexagon = Polygon2D()
    print(hexagon)
    hexagon.append(QPointF(10, 20))
    hexagon.append(QPointF(10, 30))
    hexagon.append(QPointF(20, 30))



    # hexagon.append(QPointF(10, 20))
    print(hexagon)
    print(hexagon.isEmpty())
    print(hexagon.count())
    print(hexagon.stroke)
    print(hexagon.get_stroke_size())
    # ça marche maintenant
    # print(hexagon.)
    # trop cool l'acces aux points
    for point in hexagon:
        print(point)

    print(hexagon.get_points())
    # print(hexagon.contains(10, 20))
    print(hexagon.contains(QPointF(10, 20)))
    print(hexagon.contains(QPointF(10, 21)))

    print(hexagon.isClosed()) #
    hexagon.append(QPointF(10, 20)) # closing the hexagon --> the last and first point should be the same
    print(hexagon.isClosed())  #

    # print(hexagon.translate(10, 20)) # why none ???
    # translate and so on can all be saved...

    image = QImage('./../data/handCorrection.png')
    painter = QPainter()
    painter.begin(image)
    # painter.setOpacity(0.3);
    painter.drawImage(0, 0, image)
    painter.setPen(QtCore.Qt.blue)
    painter.drawPolygon(hexagon)
    hexagon.opacity = 0.7
    painter.translate(10, 20)
    hexagon.draw(painter) # ça marche pourrait overloader ça avec du svg
    painter.end()

    # painter.save()
    # painter.setCompositionMode(QtGui.QPainter.CompositionMode_Clear)
    # painter.eraseRect(r)
    # painter.restore()

    image.save('./../trash/test_pyQT_draw.png', "PNG");

    #pas mal TODO faire une classe drawsmthg qui dessine n'importe quelle forme que l'on lui passe avec des parametres de couleur, transparence, ...

    # tt marche aps mal ça va très vite

